"""Reachy parts submodule.

Used to define all parts in Reachy:

* :py:class:`RightArm`, :py:class:`LeftArm`
* :py:class:`~reachy.parts.hand.ForceGripper`
* :py:class:`Head`
"""

from .arm import LeftArm, RightArm # noqa
from .head import Head # noqa


"""Custom errors for Reachy."""


class LuosModuleNotFoundError(Exception):
    """Luos Module missing."""

    def __init__(self, message, missing_module):
        """Define Luos Module Error.

        Args:
            message (str): custom error message
            missing_module (str): name of the missing luos module
        """
        super().__init__(message)

        self.missing_module = missing_module


class LuosGateNotFoundError(Exception):
    """Luos Gate missing."""


class CameraNotFoundError(Exception):
    """Camera missing."""

    def __init__(self, message, camera_id):
        """Define Camera missing Error.

        Args:
            message (str): custom error message
            camera_id (int): index of the missing camera
        """
        super().__init__(message)

        self.camera_id = camera_id



"""Wrapper around two RPI and the multi channel switch.

It relies on the use of the Multi_Adapter_Board_2Channel_uc444 for switching camera via the I2C and GPIO control.

See https://github.com/ArduCAM/RaspberryPi/tree/master/Multi_Camera_Adapter/Multi_Adapter_Board_2Channel_uc444

"""

import time
import cv2 as cv

from threading import Thread, Event, Lock
from gpiozero import DigitalOutputDevice
from smbus import SMBus

from ..error import CameraNotFoundError


class DualCamera(object):
    """Wrapper around two RPI and the multi channel switch."""

    def __init__(self, default_camera='right'):
        """Create a DualCamera obejct and set the default active."""
        self._pin17 = DigitalOutputDevice(17)
        self._pin4 = DigitalOutputDevice(4)
        self._bus = SMBus(1)

        self.set_active(default_camera)
        # Make sure, the camera is active before trying to access it.
        time.sleep(0.5)
        self.cap = BackgroundVideoCapture(0)

    def __exit__(self):
        """Automatically close the cam on exit."""
        self.close()

    def close(self):
        """Close the camera capture."""
        self.cap.close()
        self._pin17.close()
        self._pin4.close()
        self._bus.close()

    @property
    def active_side(self):
        """Get the active camera side."""
        return self._active

    def set_active(self, camera_side):
        """Set one of the camera active (left or right)."""
        if camera_side not in ('left', 'right'):
            raise ValueError('camera_side should be either "left" or "right"!')

        if camera_side == 'left':
            self._enable_left_camera()
        elif camera_side == 'right':
            self._enable_right_camera()

        self._active = camera_side

    def read(self):
        """Get the latest retrieved frame."""
        return self.cap.read()

    def _enable_left_camera(self):
        """Enable the left eye camera.

        See https://github.com/ArduCAM/RaspberryPi/tree/master/Multi_Camera_Adapter/Multi_Adapter_Board_2Channel_uc444
        """
        self._bus.write_byte_data(0x70, 0, 0x01)
        self._pin17.off()
        self._pin4.off()

    def _enable_right_camera(self):
        """Enable the left eye camera.

        See https://github.com/ArduCAM/RaspberryPi/tree/master/Multi_Camera_Adapter/Multi_Adapter_Board_2Channel_uc444
        """
        self._bus.write_byte_data(0x70, 0, 0x02)
        self._pin4.on()


class BackgroundVideoCapture(object):
    """Wrapper on OpenCV VideoCapture object.

    Args:
        camera_index (int): index of the used camera (see OpenCV doc for details)
        resolution (int, int): desired resolution for the grabbed frame (the resolution must be compatible with the driver)

    Instantiating this object will automatically start the polling of image in background.

    This wrapper is reponsible for automatically polling image on the camera.
    This ensures that we can always access the most recent image.
    """

    def __init__(self, camera_index, resolution=(720, 960)):
        """Open video capture on the specified camera."""
        self.cap = cv.VideoCapture(camera_index)

        if not self.cap.isOpened():
            raise CameraNotFoundError(
                message=f'Camera {camera_index} not found!',
                camera_id=camera_index,
            )

        self.cap.set(cv.CAP_PROP_FRAME_HEIGHT, resolution[0])
        self.cap.set(cv.CAP_PROP_FRAME_WIDTH, resolution[1])

        self._lock = Lock()
        self.running = Event()

        self._img = None

        self._t = Thread(target=self._read_loop)
        self._t.daemon = True
        self._t.start()

    def close(self):
        """Stop polling image and release the Video Capture."""
        self.running.clear()

        if self._t.is_alive():
            self._t.join()

        self.cap.release()

    def _read_loop(self):
        self.running.set()

        while self.running.is_set():
            b, img = self.cap.read()

            if b:
                with self._lock:
                    self._img = img.copy()

    def read(self):
        """Retrieve the last grabbed image."""
        with self._lock:
            return self._img is not None, self._img



"""Wrapper module on top of pyluos Robot object."""
import time
import logging

from glob import glob

from pyluos import Device as LuosDevice
from pyluos.modules import DynamixelMotor

from ..error import LuosModuleNotFoundError, LuosGateNotFoundError
from .io import IO

logger = logging.getLogger(__name__)


def attempt_luos_connection(port, trials=5):
    """Try to connect to a Luos Gate."""
    io = LuosDevice(port, log_conf='')
    gate_name = io.modules[0].alias

    if trials > 0 and gate_name in ('r_right_arm', 'r_left_arm'):
        # TEMP: check if the dxl did respond
        if len(io.modules) < 3:
            io.close()
            time.sleep(0.1)
            return attempt_luos_connection(port, trials-1)

    return io


class SharedLuosIO(IO):
    """
    Abstraction class for pyluos Robot object. Create a new connection with a Luos gate.

    Args:
        luos_port (str): name of the serial port used (e.g. '/dev/ttyUSB0')

    .. note:: If a connection on the same port already exists, the same IO will be used.

    The class is reponsible for holding active connections with Luos gate. A same gate can be shared among multiple IOs.
    """

    opened_io = {}

    def __init__(self, luos_port):
        """Create a new connection with a Luos gate."""
        if luos_port not in SharedLuosIO.opened_io:
            io = attempt_luos_connection(luos_port)
            SharedLuosIO.opened_io[luos_port] = io

            logger.info('Connected to new Luos IO', extra={
                'luos_port': luos_port,
                'gate_name': io.modules[0].alias,
                'modules': [mod.alias for mod in io.modules],
            })
            # FIXME: wait for a first sync of all modules
            import time
            time.sleep(1)
        self.shared_io = SharedLuosIO.opened_io[luos_port]
        self.port = luos_port

    def __repr__(self):
        """Shared IO representation."""
        mod = [m.alias for m in self.shared_io.modules]
        return f'<SharedLuosIO "port": "{self.port}" "modules": {mod}>'

    @classmethod
    def with_gate(cls, name, port_template):
        """Open a connection on the specified Luos gate.

        Args:
            name (str): name (or alias) of the searched Luos gate.
            port_template (str): template name for the possible serial ports (e.g. '/dev/ttyUSB*')

        """
        logger.info(f'Looking for gate "{name}" on ports "{port_template}"')

        available_ports = glob(port_template)

        if len(available_ports) == 0:
            return cls(port_template)

        for p in available_ports:
            io = cls(p)

            if io.gate_name == name:
                logger.info(f'Found gate "{io.gate_name}" on port "{p}"')
                return io
        else:
            raise LuosGateNotFoundError(f'Gate "{name}" not found on ports "{port_template}"')

    @classmethod
    def close_all_cached_gates(cls):
        """Close all connections to the Luos gate."""
        for io in SharedLuosIO.opened_io.values():
            io.close()
        SharedLuosIO.opened_io.clear()

    @property
    def gate_name(self):
        """Retrieve the name of the Luos gate."""
        return self.shared_io.modules[0].alias

    def close(self):
        """Close the IO.

        .. warning:: You are responsible for handling correctly closing if you are using multiple connections on the same IO.
        """
        self.shared_io.close()
        logger.info('Luos IO connection closed', extra={
            'gate_name': self.gate_name,
            'port': self.port,
        })

    def find_module(self, module_name):
        """Retrieve a specified Luos module on the IO given its name.

        Args:
            module_name (str): name (or alias) of the researched module
        """
        try:
            return getattr(self.shared_io, module_name)
        except AttributeError:
            raise LuosModuleNotFoundError(
                message=f'Could not find module "{module_name}" on bus "{self.port}"',
                missing_module=module_name,
            )

    def find_dxl(self, dxl_name, dxl_config):
        """Retrieve a specified Dynamixel motor on the IO given its id.

        Args:
            dxl_config (dict): configuration of the searched dynamixel (including its id)
        """
        module_name = 'dxl_{}'.format(dxl_config['id'])

        m = self.find_module(module_name)

        if not isinstance(m, DynamixelMotor):
            raise LuosModuleNotFoundError(f'Wrong module type found for module "{module_name}" on bus "{self.port}"')

        return m

    def find_fan(self, fan_name):
        """Get a specific fan from its name."""
        return Fan(fan_name, self.find_module(fan_name))

    def find_orbita_disks(self):
        """Retrieve the three Luos modules controlling each Orbita disk."""
        return [
            OrbitaDisk(name, self.find_module(name))
            for name in ['disk_bottom', 'disk_middle', 'disk_top']
        ]

    def find_dual_camera(self, default_camera):
        """Retrieve a dual camera."""
        # We import DualCamera here to avoid OpenCV/smbus/gpiozero ImportError
        # if we are not using the Head part.
        from .cam import DualCamera
        return DualCamera(default_camera)


class OrbitaDisk(object):
    """Orbita Disk Wrapper around luos controlled motor module."""

    def __init__(self, name, luos_disk) -> None:
        """Create a new Orbita disk using the luos module.

        Args:
            name (str): name of the disk (e.g. "disk_bottom").
            luos_disk: controlled_motor luos module

        """
        self.name = name
        self.luos_disk = luos_disk
        self.offset = 0

    def __repr__(self) -> str:
        """Get the OrbitaDisk string representation."""
        return f'<Orbita "{self.name}" pos="{self.rot_position}>'

    def setup(self):
        """Prepare the luos disk before controlling it.

        Enable position control, retrieve position and temperature.
        """
        self.luos_disk.rot_position_mode = True
        self.luos_disk.rot_position = True
        self.luos_disk.temperature = True

    @property
    def compliant(self):
        """Get the disk compliancy."""
        return self.luos_disk.compliant

    @compliant.setter
    def compliant(self, new_compliancy):
        """Set new compliancy (stiff/compliant)."""
        self.luos_disk.compliant = new_compliancy

    @property
    def rot_position(self):
        """Get the current angle position (in deg.)."""
        return self.luos_disk.rot_position

    @property
    def target_rot_position(self):
        """Get the current target angle position (in deg.)."""
        return self.luos_disk.target_rot_position - self.offset

    @target_rot_position.setter
    def target_rot_position(self, new_pos):
        """Set a new target angle position (in deg.)."""
        self.luos_disk.target_rot_position = new_pos + self.offset

    @property
    def temperature(self):
        """Get the current motor temperature in C."""
        return self.luos_disk.temperature


class Fan(object):
    """Fan module for motor cooling."""

    def __init__(self, name, mod):
        """Create a new motor associated with its Luos module."""
        self.name = name
        self.mod = mod

    def __repr__(self):
        """Fan representation."""
        return f'<Fan "{self.name}" is "{self.status}">'

    @property
    def status(self):
        """Get the fan mode ('on' or 'off')."""
        return 'on' if self.mod.state else 'off'

    def on(self):
        """Turn the fan on."""
        self.mod.state = True

    def off(self):
        """Turn the fan off."""
        self.mod.state = False



"""Utility module to automatically merge trajectories.

You should always check the result of the combination of your trajectories before applying it to the robot.
Indeed, the code below is by no mean safe. Artefacts could notably be introduced in the process.

"""
import numpy as np


def sigmoid(x):
    """Compute sigmoid function."""
    return 1 / (1 + np.exp(-x))


def norm_sigmoid(n, r):
    """Compute a normalized sigmoid function with specified range.

    Args:
        n (int): number of samples to used
        r (float): X range [-r, r] where to compute the sigmoid
    """
    y = sigmoid(np.linspace(-r, r, n))
    y = y - y.min()
    y = y / y.max()
    return y


def _combine(m1, m2, overlap, r):
    n1, y = m1.shape
    n2 = m2.shape[0]

    w = int(np.round((n1 + n2) * overlap))
    n = n1 + n2 - w

    M1 = np.zeros((n, y))
    M1[:n1] = m1

    M2 = np.zeros((n, y))
    M2[-n2:] = m2

    a = n - n2
    b = n1

    W = np.zeros((n, 1))
    W[:a] = 1
    W[a:b, 0] = 1 - norm_sigmoid(b - a, r)
    W[b:] = 0

    M = ((M1 * W) + (M2 * (1 - W)))
    return M


def traj_as_array(traj):
    """Transform a trajectory dict into a numpy 2D array.

    Args:
        traj (dict): {motor_name: list of positions}
    """
    return np.array(list(traj.values())).T


def combine(*trajs, overlap=0.1, r=5):
    """Combine multiple trajectory into a single one.

    Args:
        trajs (list): list of trajectories (as dict)
        overlap (float): overlap recovery percentage for combining trajectories
        r (int): sigmoid slope factor used to merge two trajectories
    """
    motors = set([tuple(t.keys()) for t in trajs])
    if len(motors) != 1:
        raise ValueError('All trajs must use the same motors!')
    motors = motors.pop()

    moves = [traj_as_array(t) for t in trajs]

    m1 = moves[0]

    for m2 in moves[1:]:
        m1 = _combine(m1, m2, overlap, r)

    return {
        m: m1[:, i]
        for i, m in enumerate(motors)
    }



"""Discovery utility tools.

Provide automatic detection mechanisms to check which parts are present on your robot and their status.
.. warning:: Make sure that no other Python instance is connected to the robot before running those functions.
"""

from glob import glob

from reachy import parts
from reachy.io.luos import SharedLuosIO
from reachy.error import LuosModuleNotFoundError, LuosGateNotFoundError, CameraNotFoundError


def discover_head(luos_port='/dev/ttyUSB*'):
    """Check if an head part is connected to a Reachy."""
    part_name = 'head'

    try:
        head = parts.Head(
            io=luos_port,
        )
        head.teardown()
        status = 'ok'
        details = ''

    except LuosGateNotFoundError as e:
        status = 'missing'
        details = str(e)

    except LuosModuleNotFoundError as e:
        status = 'problem'
        details = f'module "{e.missing_module}" missing'

    except CameraNotFoundError as e:
        status = 'problem'
        details = f'camera {e.camera_id} missing'

    return {
        'part_name': part_name,
        'status': status,
        'details': details,
    }


def discover_arm(side, luos_port='/dev/ttyUSB*', hand='force_gripper'):
    """Check if an arm part is connected to a Reachy."""
    Arm = parts.LeftArm if side == 'left' else parts.RightArm

    part_name = f'{side}_arm'

    try:
        arm = Arm(
            io=luos_port,
            hand=hand,
        )
        arm.teardown()
        status = 'ok'
        details = ''

    except LuosGateNotFoundError as e:
        status = 'missing'
        details = str(e)

    except LuosModuleNotFoundError as e:
        status = 'problem'
        details = f'module "{e.missing_module}" missing'

    return {
        'part_name': part_name,
        'status': status,
        'details': details,
    }


def discover_all(luos_port='/dev/ttyUSB*'):
    """Check which part is connected to a Reachy."""
    SharedLuosIO.close_all_cached_gates()

    if len(glob(luos_port)) == 0:
        return {}

    return {
        'left_arm': discover_arm(luos_port=luos_port, side='left'),
        'head': discover_head(luos_port=luos_port),
        'right_arm': discover_arm(luos_port=luos_port, side='right'),
    }


def main():
    """Check which part is connected to a Reachy."""
    for name, part in discover_all.items():
        status = part['status']
        details = part['details']

        if status == 'ok':
            message = f'Part "{name}" was found and everything seems ok!'

        elif status == 'missing':
            message = f'Part "{name}" is missing, check connections if it should have been found.'

        elif status == 'problem':
            message = f'Part "{name}" was found but it seems there is a problem: {details}!!!'
        print(message)


if __name__ == '__main__':
    main()



"""Kinematic forward and inverse utility module.

* Allow the creation of kinematic chain (following a simplified DH notation)
* Compute the forward kinematic
* Provide optimization via scipy for inverse approximation
"""

import numpy as np

from scipy.spatial.transform import Rotation
from scipy.optimize import minimize


class Link(object):
    """Simplified Link following DH notation.

    Args:
        translation (:py:class:`~numpy.ndarray`): 3d translation from the previous link
        rotation (:py:class:`~numpy.ndarray`): 3d euler rotation (xyz) from the previous link
    """

    def __init__(self, translation, rotation, bounds):
        """Create a new link."""
        self.T = translation_matrix(translation)
        self.rotation = np.array(rotation).reshape(1, 3)
        self.bounds = bounds

    def transformation_matrix(self, theta):
        """Compute local transformation matrix for given angle.

        Args:
            theta (float): joint rotation angle (in radians)
        """
        R = np.zeros((theta.shape[0], 4, 4))
        theta = theta.reshape(1, -1)

        R[:, :3, :3] = Rotation.from_rotvec(np.dot(theta.T, self.rotation)).as_matrix()
        R[:, 3, 3] = 1
        return np.matmul(self.T, R)


class Chain(object):
    """Chain of Link forming a kinematic chain.

    Args:
        links (list): list of :py:class:`Link`
    """

    def __init__(self, links):
        """Create the chain given the links."""
        self.links = links

    @property
    def bounds(self):
        """Get bounds for each link."""
        return [link.bounds for link in self.links]

    def forward(self, joints):
        """
        Compute forward kinematics of the chain given joints configurations.

        Args:
            joints (:py:class:`~numpy.ndarray`): N*J array joint rotation angle for each link (in radians)

        Returns:
            :py:class:`~numpy.ndarray`: 4x4 homogeneous matrix pose of the end effector

        .. warning:: this is a vectorized version of the forward!
        """
        M = np.eye(4)

        for link, theta in zip(self.links, joints.T):
            M = np.matmul(M, link.transformation_matrix(theta))

        return M

    def inverse(self, poses, q0s, maxiter=10):
        """
        Approximate the inverse kinematics of the chain given end pose.

        Args:
            poses (:py:class:`~numpy.ndarray`): N*4*4 homogeneous matrix poses for the end effector
            q0s (:py:class:`~numpy.ndarray`): N*J initial joint configuration used to bootstrap the optimization
            maxiter (int): maximum number of iteration to run on the optimizer

        .. warning:: this is a vectorized version of the forward!
        """
        return np.array([
            self._inverse(p, q0, maxiter)
            for p, q0 in zip(poses, q0s)
        ])

    def _inverse(self, target, q0, maxiter):
        def forward_error(j):
            M = self.forward(np.array(j).reshape(1, -1))[0]
            d = pose_dist(M, target)

            return d

        sol = minimize(
            forward_error,
            x0=q0,
            options={
                'maxiter': maxiter,
            },
            bounds=self.bounds,
        )
        return sol.x


def translation_matrix(translation):
    """Create homogenous matrix given a 3D translation vector."""
    M = np.eye(4)
    M[:3, 3] = translation
    return M


def position_dist(P, Q):
    """Compute euclidian distance between two 3D position."""
    return np.linalg.norm(P - Q)


def rotation_dist(P, Q):
    """Compute rotation distance between two 3D rotation."""
    R = np.matmul(P, Q.T)

    A = (np.trace(R) - 1) / 2
    A = np.clip(A, -1, 1)

    theta = np.arccos(A)
    return theta


def pose_dist(M1, M2, threshold=-1):
    """Compute distance between two pose.

    The distance is defined as the sum of the position distance and the rotation distance where 1° of error is equal to 1mm error distance.
    """
    P1, R1 = M1[:3, 3], M1[:3, :3]
    P2, R2 = M2[:3, 3], M2[:3, :3]

    PD = position_dist(P1, P2)
    RD = rotation_dist(R1, R2)

    # meaning 1m pos error ~ 0.2 rad rot error
    E = PD + 0.2 * RD

    return E if E > threshold else 0



"""Motor abstraction module.

Define:
    * DynamixelMotor
    * OrbitaActuator
"""

import time
import logging
import numpy as np

from threading import Timer
from orbita import Actuator as OrbitaModel

from ..trajectory.interpolation import interpolation_modes

logger = logging.getLogger(__name__)


class DynamixelMotor(object):
    """DynamixelMotor abstraction class.

    Args:
        root_part (str): name of the part where the motor is attached to (eg 'right_arm.hand')
        name (str): name of the motor (eg. 'shoulder_pitch')
        luos_motor (:py:class:`pyluos.modules.DxlMotor`): pyluos motor
        config (dict): extra motor config (must include 'offset' and 'orientation' fields)

    Wrap the pyluos motor object to simplify and make the API homogeneous.
    """

    def __init__(self, root_part, name, luos_motor, config):
        """Create a DynamixelMotor given its pyluos equivalent."""
        self._root_part = root_part
        self._name = name

        self._motor = luos_motor

        self._offset = config['offset']
        self._direct = config['orientation'] == 'direct'

        self._timer = None
        self._use_static_fix = False

    def __repr__(self):
        """Motor representation."""
        mode = 'compliant' if self.compliant else 'stiff'
        return f'<DxlMotor "{self.name}" pos="{self.present_position}" mode="{mode}">'

    @property
    def name(self):
        """Fullname of the motor (eg. right_arm.hand.gripper)."""
        return f'{self._root_part.name}.{self._name}'

    # Position
    @property
    def present_position(self):
        """Present position (in degrees) of the motor."""
        return self._as_local_pos(self._motor.rot_position)

    @property
    def goal_position(self):
        """Get current goal position (in degrees) of the motor."""
        return self._as_local_pos(self._motor.target_rot_position)

    @goal_position.setter
    def goal_position(self, value):
        if not self.compliant:
            self._motor.target_rot_position = self._to_motor_pos(value)

            if self._use_static_fix:
                self._schedule_static_error_fix(delay=1)

    @property
    def offset(self):
        """Get motor real zero (in degrees)."""
        return self._offset

    def is_direct(self):
        """Check whether the motor is direct or not."""
        return self._direct

    def _as_local_pos(self, pos):
        return (pos if self.is_direct() else -pos) - self.offset

    def _to_motor_pos(self, pos):
        return (pos + self.offset) * (1 if self.is_direct() else -1)

    # Speed
    @property
    def moving_speed(self):
        """Get the maximum speed (in degree per second) of the motor."""
        return self._motor.target_rot_speed

    @moving_speed.setter
    def moving_speed(self, value):
        self._motor.target_rot_speed = value

    # Compliancy
    @property
    def compliant(self):
        """Check whether or not the motor is compliant."""
        return self._motor.compliant

    @compliant.setter
    def compliant(self, value):
        self._motor.compliant = value

    @property
    def torque_limit(self):
        """Check the maximum torque allowed (in %) of the motor."""
        return self._motor.power_ratio_limit

    @torque_limit.setter
    def torque_limit(self, value):
        self._motor.power_ratio_limit = value

    # Temperature
    @property
    def temperature(self):
        """Check the current motor temp. (in °C)."""
        return self._motor.temperature

    def goto(self,
             goal_position, duration,
             starting_point='present_position',
             wait=False, interpolation_mode='linear'):
        """Set trajectory goal for the motor.

        Args:
            goal_position (float): target position (in degrees)
            duration (float): duration of the movement (in seconds)
            starting_point (str): register used to determine the starting point (eg. 'goal_position' can also be used in some specific case)
            wait (bool): whether or not to wait for the end of the motion
            interpolation_mode (str): interpolation technique used for computing the trajectory ('linear', 'minjerk')

        Returns:
            reachy.trajectory.TrajectoryPlayer: trajectory player that can be used to monitor the trajectory, stop it, etc
        """
        if interpolation_mode not in interpolation_modes.keys():
            available = tuple(interpolation_modes.keys())
            raise ValueError(f'interpolation_mode should be one of {available}')

        traj_player = interpolation_modes[interpolation_mode](getattr(self, starting_point), goal_position, duration)
        traj_player.start(self)

        if wait:
            traj_player.wait()

        return traj_player

    def use_static_error_fix(self, activate):
        """Trigger the static error fix.

        Args:
            activate (bool): whether to activate/deactivate the static error issue fix

        If activated, the static error fix will check the reach position a fixed delay after the send of a new goal position.
        The static error may result in the motor's load increasing, and yet not managing to move.
        To prevent this behavior we automatically adjust the target goal position to reduce this error.
        """
        self._use_static_fix = activate

    # Patch dynamixel controller issue when the motor forces
    # while not managing to reach the goal position
    def _schedule_static_error_fix(self, delay):
        if self._timer is not None:
            self._timer.cancel()
        self._timer = Timer(delay, self._fix_static_error)
        self._timer.start()

    def _fix_static_error(self, threshold=2):
        error = (self.present_position - self.goal_position)

        if abs(error) > threshold:
            pos = self.goal_position + error / 2
            logger.info('Fix static error controller', extra={
                'goal_position': self.goal_position,
                'present_position': self.present_position,
                'fixed_goal_position': pos,
            })

            self._motor.target_rot_position = self._to_motor_pos(pos)
            self._timer = None


class OrbitaActuator(object):
    """Orbita Actuator abstraction.

    Args:
        root_part (str): name of the part where the motor is attached to (eg 'head')
        name (str): name of the actuator (eg. 'neck')
        disks_motor (list of :py:class:`pyluos.motor_controller`): list of the three disks controllers
        Pc_z (float, float, float): 3D coordinates of the center of the platform (in mm)
        Cp_z (float, float, float): center of the disks rotation circle (in mm)
        R (float): radius of the arms rotation circle around the platform (in mm)
        R0 (:py:class:`~numpy.ndarray`): rotation matrix for the initial rotation
        hardware_zero (float, float, float): absolute hardware zero position of the orbita disks

    Wrap the three disk and the computation model of Orbita to expose higher-level functionalities such as:

    * quaternion control
    * compliancy mode
    * goto
    """

    def __init__(
        self, root_part, name, disks_motor,
        Pc_z, Cp_z, R, R0, hardware_zero
    ):
        """Create a OrbitaActuator given its three disks controllers."""
        self.disk_bottom, self.disk_middle, self.disk_top = disks_motor

        self.model = OrbitaModel(Pc_z=Pc_z, Cp_z=Cp_z, R=R, R0=R0)
        self._hardware_zero = hardware_zero

        self._compliancy = False
        self.setup()

    def __repr__(self):
        """Orbita representation."""
        return (f'<Orbita '
                f'"top disk": {self.disk_top.rot_position} '
                f'"middle disk": {self.disk_middle.rot_position} '
                f'"bottom disk": {self.disk_bottom.rot_position}>')

    @property
    def disks(self):
        """Get three disks [top, middle, bottom]."""
        return [self.disk_top, self.disk_middle, self.disk_bottom]

    @property
    def compliant(self):
        """Check the disks compliancy."""
        return self._compliancy

    @compliant.setter
    def compliant(self, compliancy):
        self._compliancy = compliancy

        for d in self.disks:
            d.compliant = compliancy

    def goto(self,
             thetas, duration, wait,
             interpolation_mode='linear',
             ):
        """Set trajectory goal for three disks.

        Args:
            thetas (float, float, float): target position (in degrees) for each disks (top, middle, bottom)
            duration (float): duration of the movement (in seconds)
            wait (bool): whether or not to wait for the end of the motion
            interpolation_mode (str): interpolation technique used for computing the trajectory ('linear', 'minjerk')

        Returns:
            reachy.trajectory.TrajectoryPlayer: trajectory player that can be used to monitor the trajectory, stop it, etc
        """
        if len(thetas) != len(self.disks):
            raise ValueError(f'Invalid thetas {thetas} (length should be {len(self.disks)}')

        if interpolation_mode not in interpolation_modes.keys():
            available = tuple(interpolation_modes.keys())
            raise ValueError(f'interpolation_mode should be one of {available}')
        Traj = interpolation_modes[interpolation_mode]

        trajs = [
            Traj(
                initial_position=disk.target_rot_position,
                goal_position=angle,
                duration=duration,
            )
            for disk, angle in zip(self.disks, thetas)
        ]

        for disk, traj in zip(self.disks, trajs):
            traj.start(disk)

        if wait:
            for traj in trajs:
                traj.wait()

        return trajs

    def point_at(self, vector, angle, duration, wait):
        """Make orbita point at the given vector.

        Args:
            vector (float, float, float): 3D vector indicating the poiting direction (in m)
            angle (float):  rotation around the vector (in degrees)
            duration (float): move duration (in seconds)
            wait (bool): whether or not to wait for the end of the motion

        Returns:
            reachy.trajectory.TrajectoryPlayer: trajectory player that can be used to monitor the trajectory, stop it, etc
        """
        thetas = self.model.get_angles_from_vector(vector, angle)
        # We used a reversed encoder so we need to inverse the angles
        return self.goto(thetas, duration=duration, wait=wait, interpolation_mode='minjerk')

    def orient(self, quat, duration, wait):
        """Orient orbita given a quaternion.

        Args:
            quat (pyquaternion.Quaterion): quaternion goal orientation
            duration (float): move duration (in seconds)
            wait (bool): whether or not to wait for the end of the motion

        Returns:
            reachy.trajectory.TrajectoryPlayer: trajectory player that can be used to monitor the trajectory, stop it, etc
        """
        thetas = self.model.get_angles_from_quaternion(quat.w, quat.x, quat.y, quat.z)
        # We used a reversed encoder so we need to inverse the angles
        return self.goto(thetas, duration=duration, wait=wait, interpolation_mode='minjerk')

    def setup(self):
        """Configure each of the three disks.

        .. note:: automatically called at instantiation.
        """
        for disk in self.disks:
            disk.setup()

        def _find_zero(disk, z):
            A = 360 / (52 / 24)
            p = disk.rot_position

            zeros = [z, -(A - z), A + z]
            distances = [abs(p - z) for z in zeros]
            best = np.argmin(distances)

            return zeros[best]

        time.sleep(0.25)

        for d, z in zip(self.disks, self._hardware_zero):
            d.offset = _find_zero(d, z) + 60



#!/usr/bin/env python

import re

from setuptools import setup, find_packages
from os import path


def version():
    with open('reachy/_version.py') as f:
        _, version = f.readlines()
        return re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version).group(1)


here = path.abspath(path.dirname(__file__))

with open(path.join(here, '..', 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='reachy',
    version=version(),
    packages=find_packages(exclude=['tests']),

    package_data={'reachy': ['*.npy']},

    install_requires=[
        'pyluos>=1.2.2',
        'numpy',
        'scipy>=1.4.0',
        'orbita>=0.3.1',
        'pyquaternion',
        'websockets',
        'Pillow',
    ],

    extras_require={
        'head': ['opencv-python'],
        'log': ['zzlog'],

        'doc': ['sphinx', 'sphinx-autoapi'],
        'lint': ['flake8', 'pydocstyle'],
        'test': ['pytest'],
    },

    entry_points={
        'console_scripts': [
            'dxl-config=reachy.utils.dxl_config:main',
            'orbita-config=reachy.utils.orbita_config:main',
            'reachy-setup-motorlimits=reachy.utils.setup_angle_limits:main',
            'orbita-zero=reachy.utils.orbita_zero:main',
        ],
    },

    author='Pollen Robotics',
    author_email='contact@pollen-robotics.com',
    url='https://github.com/pollen-robotics/reachy',

    description='Open source interactive robot to explore real-world applications!',
    long_description=long_description,
    long_description_content_type='text/markdown',
)



"""Part abstraction module."""

import numpy as np

from .motor import DynamixelMotor, OrbitaActuator
from .kinematic import Link, Chain

from ..io import IO, luos, ws


class ReachyPart(object):
    """Part abstraction class.

    Args:
        name (str): name of the new part, can be composed if it's a subpart (eg. right_arm.hand)
        io (str): port name where the modules can be found

    Define instantiation, teardown functionalities.
    Also provides attach function for dynamixel motors and orbita actuator.
    """

    def __init__(self, name, io):
        """Create a new part."""
        self.name = name
        self.motors = []

        if isinstance(io, IO):
            self.io = io
        elif isinstance(io, str) and io == 'ws':
            self.io = ws.WsIO.shared_server(self.name)
        else:
            gate_name = self.name.split('.')[0]
            gate_name = f'r_{gate_name}'
            self.io = luos.SharedLuosIO.with_gate(gate_name, io)

    def teardown(self):
        """Clean up before closing."""
        self.io.close()

    def attach_dxl_motors(self, dxl_motors):
        """Attach given dynamixel motors to a part.

        Args:
            dxl_motors (dict): motors config, the config must at least include an id for each motor (see attach_kinematic_chain for extra parameters)
        """
        self.motors = []

        for motor_name, config in dxl_motors.items():
            m = DynamixelMotor(self, motor_name, self.io.find_dxl(motor_name, config), config)
            setattr(self, motor_name, m)
            self.motors.append(m)

    def create_orbita_actuator(self, name, config):
        """Attach an orbita actuator to a part.

        Args:
            name (str): name of the orbita actuator (eg neck for the head part)
            config (dict): orbita configuration (see OrbitaActuator for details)
        """
        disks_motor = self.io.find_orbita_disks()
        orb = OrbitaActuator(self, name, disks_motor, **config)
        setattr(self, name, orb)
        return orb

    def attach_kinematic_chain(self, dxl_motors):
        """Attach a kinematic chain composed of the given motors.

        Args:
            dxl_motors (dict): dynamixel motors config (as given in attach_dxl_motors), the config should also include 'link-translation' and 'link-rotation' for each motor
        """
        def compute_bounds(m):
            pos = np.array(m['angle-limits'])
            bounds = (pos if m['orientation'] == 'direct' else -pos) - m['offset']
            bounds = np.deg2rad(bounds)
            lb = min(bounds)
            rb = max(bounds)

            return (lb, rb)

        self.kin_chain = Chain([
            Link(m['link-translation'], m['link-rotation'], compute_bounds(m))
            for m in dxl_motors.values()
        ])



"""Head part module."""

import os
import pathlib
import numpy as np

from collections import OrderedDict

import reachy

from ..utils import rot
from .part import ReachyPart


class Head(ReachyPart):
    """Head part.

    Args:
        io (str): port name where the modules can be found
        default_camera (str): default camera to enable ('left' or 'right')

    Composed of an orbita actuator as neck, two controlled antennas and one camera.
    """

    orbita_config = {
        'Pc_z': [0, 0, 23],
        'Cp_z': [0, 0, 0],
        'R': 35.9,
        'R0': np.dot(rot('z', 60), rot('y', 10)),
        'hardware_zero': np.load(os.path.join(pathlib.Path(reachy.__file__).parent, 'orbita_head_hardware_zero.npy')),
    }

    dxl_motors = OrderedDict([
        ('left_antenna', {
            'id': 30, 'offset': 0.0, 'orientation': 'direct',
            'angle-limits': [-150, 150],
        }),
        ('right_antenna', {
            'id': 31, 'offset': 0.0, 'orientation': 'direct',
            'angle-limits': [-150, 150],
        }),
    ])

    def __init__(self, io, default_camera='right'):
        """Create new Head part."""
        ReachyPart.__init__(self, name='head', io=io)

        self.neck = self.create_orbita_actuator('neck', Head.orbita_config)
        self.attach_dxl_motors(Head.dxl_motors)
        self.camera = self.io.find_dual_camera(default_camera)

    def __repr__(self):
        """Head representation."""
        return f'<Head "neck": {self.neck}>'

    def teardown(self):
        """Clean and close head part."""
        self.camera.close()
        ReachyPart.teardown(self)

    def look_at(self, x, y, z, duration, wait):
        """Make the head look at a 3D point in space.

        Args:
            x (float): x coordinates in space
            y (float): y coordinates in space
            z (float): z coordinates in space
            duration (float): move duration (in seconds)
            wait (bool): whether or not to wait for the end of the motion
        """
        q = self.neck.model.find_quaternion_transform([1, 0, 0], [x, y, z])
        return self.neck.orient(q, duration=duration, wait=wait)

    @property
    def compliant(self):
        """Check if the neck is compliant."""
        return self.neck.compliant

    @compliant.setter
    def compliant(self, compliancy):
        self.neck.compliant = compliancy

    @property
    def moving_speed(self):
        """Get the current disk moving speed."""
        return self.neck.moving_speed

    @moving_speed.setter
    def moving_speed(self, speed):
        self.neck.moving_speed = speed

    @property
    def active_camera(self):
        """Get the active camera side (left or right)."""
        return self.camera.active_side

    def enable_camera(self, camera_side):
        """Enable one of the camera active (left or right)."""
        self.camera.set_active(camera_side)

    def get_image(self):
        """Get last grabbed image from the camera."""
        _, img = self.camera.read()
        return img
